import torch
from functools import partial
from diisco.math_ops import rbf_kernel
from diisco.gaussian_process import GaussianProcessRegressor


class PriorFModel:
    def __init__(self, length_scale, variance, sigma_y):
        self.length_scale = length_scale
        self.variance = variance
        self.sigma_y = sigma_y

        self.kernel = partial(
            rbf_kernel, length_scale=length_scale, variance=variance
        )
        self.models = None

        self.is_fitted = False

    def fit(self, timepoints, proportions):
        """
        Fit the model to the data.
        :param timepoints: torch.tensor of shape (n_samples, 1)
        :param proportions: torch.tensor of shape (n_samples, n_proportions)
        :return:
        """
        n_samples, n_proportions = proportions.shape
        if not timepoints.shape == (n_samples, 1):
            raise ValueError(
                "The timepoints must be a column vector of shape (n_samples, 1)."
            )

        self.models = []
        for target in proportions.T:
            model = self._fit_target(timepoints, target)
            self.models.append(model)

        self.is_fitted = True

    def _fit_target(self, timepoints, target):
        """
        Fit a single target.
        :param timepoints: torch.tensor of shape (n_samples, 1)
        :param target: torch.tensor of shape (n_samples,)
        :return:
        """
        model = GaussianProcessRegressor(self.kernel, self.sigma_y)

        # Make sure that the target is a column vector
        target = target.reshape(-1, 1)
        model.fit(timepoints, target)
        return model

    def predict(self, timepoints):
        """
        Predict the latent features for the given timepoints.
        :param timepoints: torch.tensor of shape (n_samples, 1)
        :return: means, covariances of shape (n_targets, n_samples, 1)
            and (n_targets, n_samples, n_samples)
        """
        assert self.is_fitted

        n_samples, n_targets = len(timepoints), len(self.models)
        means = torch.zeros(n_targets, n_samples, 1)
        covariances = torch.zeros(n_targets, n_samples, n_samples)

        for i, model in enumerate(self.models):
            mean, covariance = model.predict(timepoints)
            means[i] = mean
            covariances[i] = covariance

        assert means.shape == (n_targets, n_samples, 1)
        assert covariances.shape == (n_targets, n_samples, n_samples)
        return means, covariances



import torch
from scipy.stats import norm


def flat_vect_to_lower_triangular_matrix(vect, n, batch_size):
    """
    Convert a flat vector to a lower triangular matrix.
    :param vect: The flat vector.
    :param n: The number of rows and columns of the matrix.
    :return: The lower triangular matrix.
    """
    assert vect.shape[0] == int(n * (n + 1) / 2) * batch_size
    # A tensor of dimensions (batch_size, n, n)
    lower_triangular_matrix = torch.zeros((batch_size, n, n))

    bool_mask = torch.tril(torch.ones((n, n))).bool()
    bool_mask = bool_mask.repeat(batch_size, 1, 1)
    lower_triangular_matrix[bool_mask] = vect
    return lower_triangular_matrix


def make_psd(mat):
    """
    Make a matrix positive semidefinite.
    :param mat: The matrix to make positive semidefinite. shape (batch_size, n, n)
    or (n, n)
    :return: The positive semidefinite matrix.
    """

    # Add a small constant to the diagonal to make sure the matrix is positive definite
    small_constant = 0.000000001

    if mat.shape[0] == mat.shape[1]:
        if len(mat.shape) == 3:
            mat = (
                mat
                + torch.eye(mat.shape[-1]).repeat(mat.shape[0], 1, 1)
                * small_constant
            )
        else:
            mat = mat + torch.eye(mat.shape[-1]) * small_constant

    return mat


def make_symmetric(mat):
    """
    Make a matrix symmetric alongside the largest square block.
    :param mat: The matrix to make symmetric. shape (batch_size, n_1, n_2)
    or (n_1, n_1)
    :return: The symmetric matrix.
    """
    if mat.shape[-1] == mat.shape[-2]:
        return (mat + mat.transpose(-1, -2)) / 2
    else:
        min_dim = min(mat.shape[-1], mat.shape[-2])
        block = mat[..., :min_dim, :min_dim]
        block = (block + block.transpose(-1, -2)) / 2
        mat[..., :min_dim, :min_dim] = block
        return mat


def shape_and_rate_from_range(
    lower_bound: float, upper_bound: float, confidence_interval: float
) -> (float, float):
    """
    Computes the shape and rate parameters of a Gamma distribution so
    that approximately the given confidence interval is covered with
    the coverage specified by the "confidence_interval" parameter.

    The function approximates a gamma with a normal distribution
    with the same mean and variance and then computes the shape and
    rate parameters of the gamma distribution that cover the same
    confidence interval as the normal distribution.
    """
    tail = (1 - confidence_interval) / 2
    c = norm.ppf(1 - tail)

    range = upper_bound - lower_bound
    mean = (upper_bound + lower_bound) / 2
    beta = 2 * c**2 * (2 * mean) / range**2
    alpha = c**2 * (2 * mean) ** 2 / range**2
    return alpha, beta



"""
Simple module for storing the names of the different strings that need to be used in
the specification of pyro models. This is done done to avoid typos and to use the
interpreter to check for errors in consistency.
"""

# Names for the different parameters of the model
LENGTHSCALE_W = "lengthscale_w"  # lengthscale for the weights
LENGTHSCALE_W_RANGE = "lengthscale_w_range"  # the range of all values of the lengthscale for the weights
LENGTHSCALE_F = "lengthscale_f"  # lengthscale for the latent features
VARIANCE_W = "variance_w"  # constant C in C * K(x, x) for the weights
VARIANCE_F = "variance_f"  # constant C in C * K(x, x) for the latent features
PRIOR_VARIANCE_F = "prior_variance_f"  # variance used for fitting prior
SIGMA_Y = "sigma_y"  # Constant C in K(x, x) + C * I i
SIGMA_W = "sigma_w"  # Constant C in K(x, x) + C * I
SIGMA_F = "sigma_f"  # Constant C in K(x, x) + C * I
W = "W"  # Weights
F = "F"  # Latent features
Y = "Y"  # Output
B = "B"  # Bias



import torch

import diisco.utils as utils


def is_psd(mat):
    is_symmetric = bool((mat == mat.T).all())
    is_positive_definite = bool(torch.all(torch.symeig(mat)[0] > 0))
    return is_symmetric and is_positive_definite


def rbf_kernel(
    x1: torch.Tensor,
    x2: torch.Tensor,
    length_scale: float,
    variance: float = 1.0,
) -> torch.Tensor:
    """
    Compute the RBF kernel between x1 and x2.
    Based on Pyro's implementation of the RBF kernel.
    :param x1: First input. shape: (n1, d)
    :param x2: Second input. shape: (n2, d)
    :param length_scale: Length scale of the kernel.
    :return: Kernel matrix. shape: (n1, n2)
    """
    x1_scaled = x1 / length_scale
    x2_scaled = x2 / length_scale
    dists = torch.cdist(x1_scaled, x2_scaled, p=2)
    covariance = variance * torch.exp(-0.5 * dists**2)
    covariance = utils.make_psd(covariance)
    return covariance



"""
Simple module containing a poorman's version of a Gaussian Process Regression.
The module is based on the implementation of the Gaussian Process Regression
as described in the book "Gaussian Processes for Machine Learning" by
Carl Edward Rasmussen and Christopher K. I. Williams. The implementation
follows Algorithm 2.1  on page 19.
book: http://gaussianprocess.org/gpml/chapters/RW.pdf
"""

import torch

from diisco.utils import make_psd


class GaussianProcessRegressor:
    def __init__(self, kernel, sigma_y=1e-6):
        self.kernel = kernel
        self.sigma_y = sigma_y
        self.noise = sigma_y**2

        self._fit = False
        self.X = None
        self.y = None
        self.L = None
        self.alpha = None

    def fit(self, X, y):
        """
        Fit the Gaussian Process Regression model to the data.
        :param X: The input data. shape (n_samples, n_features)
        :param y: The output data. shape (n_samples, n_targets)
        """
        self.X = X
        self.y = y
        K = self.kernel(X, X) + self.noise * torch.eye(X.shape[0])

        self.L = torch.linalg.cholesky(K)
        self.alpha = torch.cholesky_solve(y, self.L)  # L.T @ L @ alpha = y

    def predict(self, X):
        """
        Predict the output for the given input data.
        :param X: The input data. shape (n_samples, n_features)
        :return: The predicted output
            :mean: The mean of the predicted output. shape (n_samples, n_targets)
            :var: The variance of the predicted output. shape (n_samples, n_samples)
        """
        K_star = self.kernel(self.X, X)
        assert K_star.shape[0] == self.X.shape[0]
        K_star_t = K_star.T
        mean = K_star_t @ self.alpha
        assert mean.shape[0] == X.shape[0]
        v = torch.cholesky_solve(K_star, self.L)
        var = self.kernel(X, X) - K_star_t @ v
        var = make_psd(var)

        assert mean.shape == (X.shape[0], self.y.shape[1])
        assert var.shape == (X.shape[0], X.shape[0])
        return mean, var

    def sample(self, X, n_samples=1, with_cov=False):
        """
        Sample the output for the given input data.
        :param X: The input data. shape (n_samples, n_features)
        :param n_samples: The number of samples to draw. Default is 1.
        :return: The sampled output. shape (n_samples, n_targets, dim_targets)
        """
        mean, var = self.predict(X)
        assert mean.shape == (X.shape[0], self.y.shape[1])
        assert var.shape == (X.shape[0], X.shape[0])
        if not with_cov:
            dig_var = torch.diag(var)
            var = torch.diag(dig_var)
            var.clamp_(min=0)
        var = var + self.noise * torch.eye(var.shape[0])
        samples = torch.distributions.MultivariateNormal(
            mean.squeeze(), var
        ).sample((n_samples,))
        samples = samples.view(n_samples, X.shape[0], self.y.shape[1])
        assert samples.shape == (
            n_samples,
            X.shape[0],
            self.y.shape[1],
        ), samples.shape
        return samples



"""
Contains the main class for the DIISCO algorithm.
"""


from functools import partial
from typing import Callable, Dict
from collections import defaultdict

import pyro
import pyro.distributions as dist
import torch
from pyro import poutine
from pyro.distributions import constraints
from tqdm import tqdm

import diisco.constants as constants
import diisco.names as names
import diisco.guides as guides
from diisco.gaussian_process import GaussianProcessRegressor
from diisco.math_ops import rbf_kernel
from diisco.utils import shape_and_rate_from_range
from diisco.prior_f_model import PriorFModel


class DIISCO:
    default_hypers_to_optim = [
        names.LENGTHSCALE_W,
    ]
    all_hypers = [
        names.SIGMA_Y,
        names.SIGMA_W,
        names.SIGMA_F,
        names.LENGTHSCALE_W,
        names.LENGTHSCALE_F,
        names.VARIANCE_W,
        names.VARIANCE_F,
    ]

    def __init__(
        self,
        lambda_matrix: torch.Tensor,
        hypers_init_vals: dict = None,
        use_bias: bool = True,
        verbose: bool = False,
        verbose_freq: int = 100,
    ):
        """
        Initialize the DIISCO model.
        :param lambda_matrix: A matrix of shape (n_cell_types, n_cell_types) that
            contains the values used for regularization in the kernel matrix.
        :param use_bias: Whether to use a bias term in the model. The bias term
            will have the hyper-parameters as those used for the W matrix.
        :param hypers_init_vals: A dictionary that contains the initial values for
            the hyperparameters. If any hyperparameter is not specified the default
            value will be used. See the constants module for default value and valid
            hyperparameter names.
        :param verbose: Whether to print the loss during training.
        :param verbose_freq: How often to print the loss during the training run.
        """
        self.model = None
        self.guide = None
        self.svi = None
        self.losses = []
        self.device = None  # Currently not used

        self.lambda_matrix = lambda_matrix
        self.use_bias = use_bias

        self.train_timepoints = None
        self.train_proportions = None

        self._check_user_hypers_init_vals(hypers_init_vals)
        self.hypers_init_vals = self._get_full_hypers_init_vals(
            hypers_init_vals
        )

        self.prior_model = None

        self._prior_is_set = False  # Flag to check if the prior has been set
        self._is_fit = False  # Flag to check if the model has been fit
        self.prior_f_mean = None  # The means to use for the emission model
        self.prior_f_cov = (
            None  # The covariance to use for the variational distribution
        )

        self.verbose = verbose
        self.verbose_freq = verbose_freq
        self.sub_size = None

    def fit(
        self,
        timepoints: torch.Tensor,
        proportions: torch.Tensor,
        n_iter: int = 1000,
        lr: float = 0.1,
        guide: str = "MultivariateNormal",
        subsample_size: int = None,
        hypers_to_optim: dict = None,
    ):
        """
        This function fits the DIISCO model fully to the data. It is
        equivalent to normalizing, setting the prior, and then running variational
        inference on the model.
        :param timepoints: A tensor of shape (n_timepoints, 1) that contains
            the timepoints at which the cell proportions were measured.
        :param proportions: A tensor of shape (n_timepoints, n_cell_types) that
             contains the cell proportions at each timepoint.
        :param n_iter: The number of iterations to run variational inference for.
        :param lr: The learning rate to use for variational inference.
        :param guide: The guide to use for variational inference. Currently only
            "MultivariateNormal" and "MultivariateNormalFactorized" are supported.
        :param hypers_to_optimize: A list of hyperparameters to optimize to optimize
            using variational inference. This feature is very experimental and it
            probably best left unused.
        """
        self.subsample_size = subsample_size

        self.train_timepoints = timepoints
        self.train_proportions = proportions

        self.n_cell_types = proportions.shape[1]
        self.n_timepoints = proportions.shape[0]

        pyro.clear_param_store()

        self.fit_and_set_f_prior_params(
            timepoints=timepoints,
            proportions=proportions,
            hypers=self.hypers_init_vals,
        )
        self.n_cell_types = proportions.shape[1]
        self.n_timepoints = proportions.shape[0]

        hypers_to_optim = hypers_to_optim or constants.DEFAULT_HYPERS_TO_OPTIM
        all_hypers = constants.DEFAULT_HYPERS.keys()
        hypers_to_block = [h for h in all_hypers if h not in hypers_to_optim]
        self.model = pyro.poutine.block(self._model, hide=hypers_to_block)

        self.guide = self._get_guide(guide, self.model)
        self.svi = pyro.infer.SVI(
            model=self.model,
            guide=self.guide,
            optim=pyro.optim.Adam({"lr": lr}),
            loss=pyro.infer.Trace_ELBO(),
        )

        self.losses = []
        self._is_fit = True
        for i in range(n_iter):
            loss = self.svi.step(timepoints, proportions)
            loss = (
                loss / self.n_timepoints * self.n_cell_types * self.n_cell_types
            )
            self.losses.append(loss)
            if self.verbose and i % self.verbose_freq == 0:
                print("[iteration %04d] loss: %.4f" % (i + 1, loss))

    def _get_model_hypers(self) -> dict:
        """
        Runs the sample sites associated with the model hyperparameters
        and returns a dictionary with the values of the hyperparameters.

        Importantly, this function is effectful.
        """
        sigma_w_init = torch.tensor(self.hypers_init_vals[names.SIGMA_W])
        sigma_w = pyro.param(
            names.SIGMA_W, sigma_w_init, constraint=constraints.positive
        )
        sigma_y_init = torch.tensor(self.hypers_init_vals[names.SIGMA_Y])
        sigma_y = pyro.param(
            names.SIGMA_Y, sigma_y_init, constraint=constraints.positive
        )
        w_variance_init = torch.tensor(self.hypers_init_vals[names.VARIANCE_W])
        w_variance = pyro.param(
            names.VARIANCE_W, w_variance_init, constraint=constraints.positive
        )

        w_lenghtscale_range = self.hypers_init_vals[names.LENGTHSCALE_W_RANGE]
        w_lengthscale_mean = self.hypers_init_vals[names.LENGTHSCALE_W]
        # We will sample from a gamma and use a normal approximation
        # so that 90% of the samples are within the range.
        upper_bound = w_lengthscale_mean + w_lenghtscale_range / 2
        lower_bound = w_lengthscale_mean - w_lenghtscale_range / 2
        alpha, beta = shape_and_rate_from_range(lower_bound, upper_bound, 0.9)
        w_length_scale = pyro.sample(names.LENGTHSCALE_W, dist.Gamma(alpha, beta))

        return {
            names.SIGMA_W: sigma_w,
            names.SIGMA_Y: sigma_y,
            names.LENGTHSCALE_W: w_length_scale,
            names.VARIANCE_W: w_variance,
        }

    def _model(self, timepoints, proportions=None):
        """
        Model for the DIISCO algorithm.
        :param timepoints: Tensor of timepoints for the samples.
             shape: (n_timepoints, 1)
        :param proportions: Tensor of proportions for the samples.
            shape: (n_timepoints, n_cell_types)
        """
        # We use a dictionary to store the parameters of the model
        # so that they can be easily accessed later as a return value.
        n_cell_types = self.n_cell_types
        n_timepoints = self.n_timepoints

        hypers = self._get_model_hypers()

        sigma_w = hypers[names.SIGMA_W]
        sigma_y = hypers[names.SIGMA_Y]
        lengthscale_w = hypers[names.LENGTHSCALE_W]
        variance_w = hypers[names.VARIANCE_W]
        w_covariance = rbf_kernel(
            timepoints,
            timepoints,
            length_scale=lengthscale_w,
            variance=variance_w,
        )
        f_covariance = self.prior_f_cov
        f_mean = self.prior_f_mean

        with pyro.plate("cell_types_outer_W", n_cell_types, dim=-2):
            with pyro.plate("cell_types_inner_W", n_cell_types, dim=-1):
                w_covariance = (
                    w_covariance + torch.eye(n_timepoints) * sigma_w**2
                )
                W = pyro.sample(
                    names.W,
                    dist.MultivariateNormal(
                        torch.zeros(n_timepoints), w_covariance
                    ),
                    infer=dict(
                        baseline={
                            "use_decaying_avg_baseline": True,
                            "baseline_beta": 0.95,
                        }
                    ),
                )
                assert W.shape == (n_cell_types, n_cell_types, n_timepoints)
        assert self.lambda_matrix.shape == (n_cell_types, n_cell_types)
        W = W * self.lambda_matrix.unsqueeze(-1)

        if self.use_bias:
            b_covariance = rbf_kernel(
                timepoints,
                timepoints,
                length_scale=lengthscale_w,
                variance=variance_w,
            )
            b_covariance = b_covariance + torch.eye(n_timepoints) * sigma_w**2
            with pyro.plate("cell_types_outer_B", n_cell_types, dim=-2):
                with pyro.plate("single_inner_B", 1, dim=-1):
                    B = pyro.sample(
                        names.B,
                        dist.MultivariateNormal(
                            torch.zeros(n_timepoints), b_covariance
                        ),
                    )
                    assert B.shape == (n_cell_types, 1, n_timepoints)

        with pyro.plate("node_plate", n_cell_types, dim=-1):
            assert f_mean.shape == (n_cell_types, n_timepoints, 1)
            assert f_covariance.shape == (
                n_cell_types,
                n_timepoints,
                n_timepoints,
            )
            f_mean = f_mean.squeeze(-1)
            f = pyro.sample(
                names.F,
                dist.MultivariateNormal(f_mean, f_covariance),
                infer=dict(
                    baseline={
                        "use_decaying_avg_baseline": True,
                        "baseline_beta": 0.95,
                    }
                ),
            )
            assert f.shape == (n_cell_types, n_timepoints)

        subsample_size = self.subsample_size if self.subsample_size else None
        subsample_size = subsample_size if proportions is not None else None

        with pyro.plate(
            "data_plate_outer",
            n_timepoints,
            dim=-2,
            subsample_size=subsample_size,
        ) as ind:
            with pyro.plate("data_plate_inner", n_cell_types, dim=-1):
                W = W.permute(
                    2, 0, 1
                )  # reorder for batch matrix multiplication
                f = f.permute(1, 0)  # reorder for batch matrix multiplication
                means = torch.bmm(W, f.unsqueeze(-1)).squeeze(-1)
                assert means.shape == (n_timepoints, n_cell_types)

                if self.use_bias:
                    B = B.permute(2, 0, 1).squeeze(-1)
                    means = means + B

                if subsample_size:
                    proportions = proportions[ind]
                    means = means[ind]
                y = pyro.sample(
                    names.Y, dist.Normal(means, sigma_y), obs=proportions
                )

    def score(
        self,
        timepoints: torch.Tensor,
        proportions: torch.Tensor,
        samples: int = 1000,
    ) -> float:
        """
        Score the model on the test data. Higher is better.
        :param timepoints: Tensor of timepoints for the samples.
            shape: (n_timepoints, 1)
        :param proportions: Tensor of proportions for the samples.
            shape: (n_timepoints, n_cell_types)
        """
        n_cell_types = self.n_cell_types
        n_timepoints = timepoints.shape[0]
        assert timepoints.shape == (n_timepoints, 1)
        assert proportions.shape == (n_timepoints, n_cell_types)

        samples = self.sample(timepoints, samples)
        proportions = proportions.unsqueeze(0)

        return torch.mean(
            torch.sum((samples - proportions) ** 2, dim=-1)
        ).item()

    def sample_observed_proportions(
        self, n_samples: int = 1000
    ) -> torch.Tensor:
        """
        Sample proportions to the model at the observed timepoints.
        To do this the model uses exclusively the parameters learned
        from variational inference and does not use a GP to interpolate.
        """
        y = []
        for _ in range(n_samples):
            guide_trace = poutine.trace(self.guide).get_trace(
                self.train_timepoints
            )
            trained_model = poutine.replay(self.model, trace=guide_trace)
            trained_model_trace = poutine.trace(trained_model).get_trace(
                self.train_timepoints
            )
            y.append(trained_model_trace.nodes[names.Y]["value"])
        return torch.stack(y)

    def sample_f_prior(
        self, timepoints: torch.Tensor, n_samples: int = 1000
    ) -> torch.Tensor:
        """
        We sample f from the prior models learned from the data.
        :param timepoints: Tensor of timepoints for the samples.
            shape: (n_timepoints, 1)
        :param n_samples: Number of samples to draw from the prior.

        :return: Tensor of shape (n_samples, n_cell_types, n_timepoints)
        """
        if self.prior_f_mean is None or self.prior_f_cov is None:
            raise ValueError("The prior model must be fit before sampling.")

        n_cell_types, _, _ = self.prior_f_mean.shape
        n_timepoints = timepoints.shape[0]

        mean, cov = self.prior_model.predict(timepoints)
        distribution = dist.MultivariateNormal(mean.squeeze(-1), cov)
        samples = distribution.sample((n_samples,))
        assert samples.shape == (n_samples, n_cell_types, n_timepoints)
        return samples

    def sample(
        self,
        timepoints: torch.Tensor,
        n_samples: int = 1000,
        n_samples_per_latent: int = 1,
        include_emission_variance: bool = False,
    ) -> Dict[str, torch.Tensor]:
        """
        Predict the proportions for the given timepoints.
        :param timepoints: Tensor of timepoints for the samples.
            shape: (n_timepoints, 1)
        :param n_samples: Number of samples to draw from the posterior.
        :param sample_reuse_rate: Number of times to reuse the same
            latent sample. This is helpful for speeding up the sampling
            process. The higher the number the faster the sampling but
            the less accurate the results.
        Returns:
            A  dictionary with the samples for each parameter. The keys are:
            names.W, names.F, names.Y and the values are tensors of shape
                (n_samples, n_timepoints, n_cell_types, n_cell_types),
                (n_samples, n_timepoints, n_cell_types, 1)
                (n_samples, n_timepoints, n_cell_types)
                respectively.
        """
        # This function samples the steps from the posterior according to the
        # following algorithm (see the paper for more details):
        # 1. Use the variational family to sample W, B, and F at observed timepoints
        # 2. Use the GP to obtain samples P(W_unobserved | W_observed) and P(B_unobserved | B_observed)
        # 3. Use a GP to obtain samples P(F_unobserved | F_observed). Importantly, this GP
        #    has a mean and covariance that is not zero (because we originally condition it on the observed
        #    samples so the new prior is equal to the original posterior). To achieve the same effect we sample
        #    P(F_unobserved | F_latent_observed, F_real_observed) and then we sample F_latent_unobserved.
        # 4. Use the sampled W, B, and F to sample Y.
        # 5. Repeat steps 2-4 n_samples_per_latent times.
        samples = defaultdict(list)
        pbar = tqdm(range(n_samples))

        samples_so_far = 0
        while samples_so_far < n_samples:
            guide_trace = poutine.trace(self.guide).get_trace(
                self.train_timepoints
            )
            trained_model = poutine.replay(self._model, trace=guide_trace)
            model_trace = poutine.trace(trained_model).get_trace(
                self.train_timepoints
            )

            W_sampled = model_trace.nodes[names.W]["value"]
            f_sampled = model_trace.nodes[names.F]["value"]
            B_sampled = (
                model_trace.nodes[names.B]["value"] if self.use_bias else None
            )

            w_length_scale = model_trace.nodes[names.LENGTHSCALE_W]["value"]
            w_variance = model_trace.nodes[names.VARIANCE_W]["value"]

            sigma_y = model_trace.nodes[names.SIGMA_Y]["value"]
            sigma_w = model_trace.nodes[names.SIGMA_W]["value"]

            if not include_emission_variance:
                sigma_y = torch.ones_like(sigma_y) * 0.001

            assert W_sampled.shape == (
                self.n_cell_types,
                self.n_cell_types,
                self.n_timepoints,
            )
            assert f_sampled.shape == (self.n_cell_types, self.n_timepoints)
            if self.use_bias:
                assert B_sampled.shape == (
                    self.n_cell_types,
                    1,
                    self.n_timepoints,
                )

            W_predict = torch.zeros(
                (
                    n_samples_per_latent,
                    self.n_cell_types,
                    self.n_cell_types,
                    timepoints.shape[0],
                )
            )
            # Compute P(W_unobserved | W_observed)
            for i in range(self.n_cell_types):
                for j in range(self.n_cell_types):
                    gpr_w_kernel = partial(
                        rbf_kernel,
                        length_scale=w_length_scale,
                        variance=w_variance,
                    )

                    regressor = GaussianProcessRegressor(
                        gpr_w_kernel, sigma_y=sigma_w
                    )
                    train_timepoints = self.train_timepoints.view(-1, 1)
                    train_w = W_sampled[i, j, :].view(-1, 1)
                    timepoints = timepoints.view(-1, 1)

                    assert train_timepoints.shape == (self.n_timepoints, 1)
                    assert train_w.shape == (self.n_timepoints, 1)
                    regressor.fit(train_timepoints, train_w)
                    w_sample = regressor.sample(
                        timepoints, n_samples=n_samples_per_latent
                    )
                    assert w_sample.shape == (
                        n_samples_per_latent,
                        timepoints.shape[0],
                        1,
                    ), w_sample.shape

                    W_predict[:, i, j, :] = w_sample.squeeze(-1)

            # Compute P(B_unobserved | B_observed)
            if self.use_bias:
                B_predict = torch.zeros(
                    (
                        n_samples_per_latent,
                        self.n_cell_types,
                        1,
                        timepoints.shape[0],
                    )
                )
                for i in range(self.n_cell_types):
                    gpr_b_kernel = partial(
                        rbf_kernel,
                        length_scale=w_length_scale,
                        variance=w_variance,
                    )

                    regressor = GaussianProcessRegressor(
                        gpr_b_kernel, sigma_y=sigma_w
                    )
                    train_timepoints = self.train_timepoints.view(-1, 1)
                    train_b = B_sampled[i, :, :].view(-1, 1)
                    timepoints = timepoints.view(-1, 1)

                    assert train_timepoints.shape == (self.n_timepoints, 1)
                    assert train_b.shape == (self.n_timepoints, 1)
                    regressor.fit(train_timepoints, train_b)
                    b_sample = regressor.sample(
                        timepoints, n_samples=n_samples_per_latent
                    )
                    assert b_sample.shape == (
                        n_samples_per_latent,
                        timepoints.shape[0],
                        1,
                    ), b_sample.shape
                    B_predict[:, i, :, :] = b_sample.squeeze(-1)

            # Sample the fs at the unobserved timepoints by updating the prior
            all_train_timepoints = torch.cat(
                [self.train_timepoints, self.train_timepoints],
                dim=0,
            )
            all_observed_proportions = torch.cat(
                [self.train_proportions, f_sampled.T],
                dim=0,
            )

            prior_updated_model = PriorFModel(
                self.hypers_init_vals[names.LENGTHSCALE_F],
                self.hypers_init_vals[names.VARIANCE_F],
                self.hypers_init_vals[names.SIGMA_F],
            )
            prior_updated_model.fit(
                all_train_timepoints,
                all_observed_proportions,
            )
            f_mean, f_cov = prior_updated_model.predict(timepoints)
            delta_diag = torch.eye(timepoints.shape[0]).unsqueeze(0)
            f_cov = (
                # f_cov + delta_diag * self.hypers_init_vals[names.SIGMA_F] ** 2
                f_cov
                + delta_diag * 0.01
            )
            f_predict = dist.MultivariateNormal(
                f_mean.squeeze(-1), f_cov
            ).sample((n_samples_per_latent,))
            assert f_predict.shape == (
                n_samples_per_latent,
                self.n_cell_types,
                timepoints.shape[0],
            )

            W_predict = W_predict * self.lambda_matrix.unsqueeze(-1)
            W_predict = W_predict.permute(0, 3, 1, 2)
            assert W_predict.shape == (
                n_samples_per_latent,
                timepoints.shape[0],
                self.n_cell_types,
                self.n_cell_types,
            )
            f_predict = f_predict.permute(0, 2, 1).unsqueeze(-1)
            assert f_predict.shape == (
                n_samples_per_latent,
                timepoints.shape[0],
                self.n_cell_types,
                1,
            ), "f_predict.shape: {}".format(f_predict.shape)

            if self.use_bias:
                B_predict = B_predict.permute(0, 3, 1, 2)
                assert B_predict.shape == (
                    n_samples_per_latent,
                    timepoints.shape[0],
                    self.n_cell_types,
                    1,
                ), "B_predict.shape: {}".format(B_predict.shape)

            y_predict_mean = torch.matmul(W_predict, f_predict).squeeze(-1)
            y_predict_mean = (
                y_predict_mean + B_predict if self.use_bias else y_predict_mean
            )
            y_predict_cov = torch.eye(self.n_cell_types) * sigma_y**2
            y_predict_cov = y_predict_cov.unsqueeze(0)

            y_sampled = dist.MultivariateNormal(
                y_predict_mean,
                y_predict_cov,
            ).sample()

            samples[names.W].append(W_predict)
            samples[names.F].append(f_predict)
            samples[names.Y].append(y_sampled)
            if self.use_bias:
                samples[names.B].append(B_predict)

            # Update the progress bar
            samples_so_far += n_samples_per_latent
            pbar.update(n_samples_per_latent)
        pbar.close()

        # Convert the samples to tensors and get rid of
        # the extra samples
        for key in samples:
            samples[key] = torch.cat(samples[key], dim=0)

        return samples

    def fit_and_set_f_prior_params(
        self,
        timepoints: torch.Tensor,
        proportions: torch.Tensor,
        hypers: dict,
    ) -> None:
        """
        Set the prior for the model for f by fitting a GP to the
        proportions.

        :param timepoints: Tensor of timepoints for the samples.
            shape: (n_timepoints, 1)
        :param proportions: Tensor of proportions for the samples.
            shape: (n_timepoints, n_cell_types)
        """
        length_scale = hypers[names.LENGTHSCALE_F]
        variance = hypers[names.VARIANCE_F]
        sigma_f = hypers[names.SIGMA_F]
        self.prior_model = PriorFModel(
            length_scale=length_scale, variance=variance, sigma_y=sigma_f
        )
        self.prior_model.fit(timepoints, proportions)

        self.prior_f_mean, self.prior_f_cov = self.prior_model.predict(
            timepoints
        )
        self._prior_is_set = True

    def get_means(self, timepoints: torch.Tensor) -> Dict[str, torch.Tensor]:
        n_train_timepoints = self.train_timepoints.shape[0]
        n_cell_types = self.n_cell_types

        guide_trace = poutine.trace(self.guide).get_trace(self.train_timepoints)
        trained_model = poutine.replay(self._model, trace=guide_trace)
        model_trace = poutine.trace(trained_model).get_trace(
            self.train_timepoints
        )

        W_sampled = self.guide.W_mean
        assert W_sampled.shape == (
            n_train_timepoints,
            n_cell_types,
            n_cell_types,
        )

        f_sampled = self.guide.F_mean
        assert f_sampled.shape == (n_train_timepoints, n_cell_types, 1)

        w_length_scale = model_trace.nodes[names.LENGTHSCALE_W]["value"]
        w_variance = model_trace.nodes[names.VARIANCE_W]["value"]

        sigma_w = model_trace.nodes[names.SIGMA_W]["value"]

        W_predict = torch.zeros(
            (timepoints.shape[0], n_cell_types, n_cell_types)
        )
        for i in range(self.n_cell_types):
            for j in range(self.n_cell_types):
                gpr_w_kernel = partial(
                    rbf_kernel,
                    length_scale=w_length_scale,
                    variance=w_variance,
                )

                regressor = GaussianProcessRegressor(
                    gpr_w_kernel, sigma_y=sigma_w
                )
                train_timepoints = self.train_timepoints.view(-1, 1)
                train_w = W_sampled[:, i, j].view(-1, 1)
                timepoints = timepoints.view(-1, 1)

                assert train_timepoints.shape == (self.n_timepoints, 1)
                assert train_w.shape == (self.n_timepoints, 1)
                regressor.fit(train_timepoints, train_w)
                w_predicted_mean, _ = regressor.predict(timepoints)
                assert w_predicted_mean.shape == (timepoints.shape[0], 1)

                W_predict[:, i, j] = w_predicted_mean.squeeze()

        # Compute P(B_unobserved | B_observed)
        B_predict = torch.zeros((timepoints.shape[0], n_cell_types, 1))
        if self.use_bias:
            for i in range(self.n_cell_types):
                gpr_b_kernel = partial(
                    rbf_kernel,
                    length_scale=w_length_scale,
                    variance=w_variance,
                )

                regressor = GaussianProcessRegressor(
                    gpr_b_kernel, sigma_y=sigma_w
                )
                train_timepoints = self.train_timepoints.view(-1, 1)
                B_sampled = self.guide.B_mean
                train_b = B_sampled[:, i].view(-1, 1)
                timepoints = timepoints.view(-1, 1)

                assert train_timepoints.shape == (self.n_timepoints, 1)
                assert train_b.shape == (self.n_timepoints, 1)
                regressor.fit(train_timepoints, train_b)
                b_predicted_mean, _ = regressor.predict(timepoints)
                assert b_predicted_mean.shape == (timepoints.shape[0], 1)

                B_predict[:, i] = b_predicted_mean

        # Sample the fs at the unobserved timepoints by updating the prior
        all_train_timepoints = torch.cat(
            [self.train_timepoints, self.train_timepoints],
            dim=0,
        )

        all_observed_proportions = torch.cat(
            [self.train_proportions, f_sampled.squeeze(-1)],
            dim=0,
        )

        prior_updated_model = PriorFModel(
            self.hypers_init_vals[names.LENGTHSCALE_F],
            self.hypers_init_vals[names.VARIANCE_F],
            self.hypers_init_vals[names.SIGMA_F],
        )
        prior_updated_model.fit(
            all_train_timepoints,
            all_observed_proportions,
        )
        f_predict, _ = prior_updated_model.predict(timepoints)
        f_predict = f_predict.permute(1, 0, 2)

        W_predict = W_predict * self.lambda_matrix.unsqueeze(0)
        y_predict_mean = torch.matmul(W_predict, f_predict).squeeze(-1)
        if self.use_bias:
            print("B_predict.shape: {}".format(B_predict.shape))
            print("y_predict_mean.shape: {}".format(y_predict_mean.shape))
            y_predict_mean = y_predict_mean + B_predict.squeeze(-1)
        means = {
            names.W: W_predict.detach(),
            names.F: f_predict.squeeze(-1).detach(),
            names.Y: y_predict_mean.detach(),
            names.B: B_predict.detach() if self.use_bias else None,
        }
        return means

    def get_W_mean(self, n_samples: int = 1) -> torch.Tensor:
        """
        Estimate the mean of W by sampling from the posterior.
        Returned tensor:
            shape:(n_cell_types, n_cell_types, n_timepoints)
        """
        W_samples = []
        for _ in range(n_samples):
            guide_trace = poutine.trace(self.guide).get_trace(
                self.train_timepoints
            )
            trained_model = poutine.replay(self._model, trace=guide_trace)
            model_trace = poutine.trace(trained_model).get_trace(
                self.train_timepoints
            )
            W_sampled = model_trace.nodes[names.W]["value"]
            W_sampled = W_sampled * self.lambda_matrix.unsqueeze(-1)
            W_samples.append(W_sampled)

        W_mean = torch.stack(W_samples).mean(0)
        W_mean = W_mean.permute(2, 0, 1)
        assert W_mean.shape == (
            self.n_timepoints,
            self.n_cell_types,
            self.n_cell_types,
        )
        return W_mean

    def get_f_mean(self, n_samples: int = 1) -> torch.Tensor:

        f_samples = []
        for _ in range(n_samples):
            guide_trace = poutine.trace(self.guide).get_trace(
                self.train_timepoints
            )
            trained_model = poutine.replay(self._model, trace=guide_trace)
            model_trace = poutine.trace(trained_model).get_trace(
                self.train_timepoints
            )
            f_sampled = model_trace.nodes[names.F]["value"]
            f_samples.append(f_sampled)

        f_mean = torch.stack(f_samples).mean(0)
        f_mean = f_mean.permute(1, 0)
        assert f_mean.shape == (self.n_timepoints, self.n_cell_types)
        return f_mean

    def _check_user_hypers_init_vals(self, hypers_init_vals: dict) -> None:
        """
        Makes sure that only the hyperparameters that are specified in the model are
        passed to the model.
        """
        # Check that a dictionary is passed
        if not isinstance(hypers_init_vals, dict):
            raise TypeError("hypers_init_vals must be a dictionary.")

        # Check that the keys in the dictionary are valid
        keys_in_model = set(constants.DEFAULT_HYPERS.keys())
        keys_in_init_vals = set(hypers_init_vals.keys())

        if not keys_in_init_vals.issubset(keys_in_model):
            raise ValueError(
                "The keys in hypers_init_vals must be a subset of the keys in the model."
            )

    def _get_full_hypers_init_vals(self, hypers_init_vals: dict) -> dict:
        """
        Returns a dictionary with the hyperparameters that are specified in the model.
        If the hyperparameters are not specified in the model, they are set to None.
        """
        hypers = {}
        for key in constants.DEFAULT_HYPERS.keys():
            default_val = constants.DEFAULT_HYPERS[key]
            hypers[key] = hypers_init_vals.get(key, default_val)
        return hypers

    def _get_guide(self, guide_name: str, model: Callable) -> Callable:
        """
        Returns the guide function corresponding to the name passed.
        """
        if guide_name == "AutoNormal":
            return pyro.infer.autoguide.AutoDiagonalNormal(model)
        if guide_name == "MultivariateNormal":
            return pyro.infer.autoguide.AutoMultivariateNormal(model)
        elif guide_name == "MultivariateNormalFactorized":
            return guides.MultivariateNormalFactorized(
                self.n_cell_types,
                self.n_timepoints,
                self.prior_f_mean,
                use_bias=self.use_bias,
                lambda_matrix=self.lambda_matrix,
                hypers_init_vals=self.hypers_init_vals,
                timepoints=self.train_timepoints,
                proportions=self.train_proportions,
            )
        elif guide_name == "DiagonalNormal":
            return guides.DiagonalNormal(
                self.n_cell_types,
                self.n_timepoints,
                self.prior_f_mean,
                use_bias=self.use_bias,
                lambda_matrix=self.lambda_matrix,
                hypers_init_vals=self.hypers_init_vals,
                timepoints=self.train_timepoints,
                proportions=self.train_proportions,
            )
        else:
            raise ValueError(
                "The guide must be either 'MultivariateNormal' or 'MultivariateNormalFactorized'."
            )



"""
This module contains various constants used throughout the project
such as default values for hyperparameters, etc.
"""
import diisco.names as names

# This specifies both the default values for the hyperparameters
# and the valid hyperparameter names.

DEFAULT_HYPERS = {
    names.LENGTHSCALE_W: 100,
    names.LENGTHSCALE_W_RANGE: 3,
    names.LENGTHSCALE_F: 100,
    names.VARIANCE_W: 1.0,
    names.VARIANCE_F: 1.0,
    names.SIGMA_Y: 0.2,
    names.SIGMA_W: 0.1,
    names.SIGMA_F: 0.2,
    names.PRIOR_VARIANCE_F: 1.0,
}

DEFAULT_HYPERS_TO_OPTIM = []

# This is a constant used for numerical stability.
EPSILON = 1e-10



"""
This module implements the abstract base class for a Guide
(Variational Distribution) and particular instantiations of it.

Unfortunately, due the requirement of pyro that every sample site
in the model must be matched by a sample site in the guide, this
module is tightly coupled with the model module and both should
be examined in tandem.

See Guide base class for interface.
"""
import torch
import pyro
import pyro.distributions as dist
from pyro.distributions import constraints
from torch.distributions.utils import vec_to_tril_matrix
from sklearn.linear_model import LinearRegression

import diisco.names as names

from diisco.constants import EPSILON

GUIDE_REGISTRY = {}


def register_guide(name):
    """
    Decorator to register a new guide.
    """

    def register_guide_cls(cls):
        if name in GUIDE_REGISTRY:
            raise ValueError(
                "Cannot register duplicate guide ({})".format(name)
            )
        if not issubclass(cls, Guide):
            raise ValueError(
                "Guide ({}: {}) must extend the Guide class".format(
                    name, cls.__name__
                )
            )
        GUIDE_REGISTRY[name] = cls
        return cls

    return register_guide_cls


class Guide:
    def __init__(self, *args, **kwargs):
        pass

    def __call__(self, *args, **kwargs):
        raise NotImplementedError

    @property
    def W_mean(self):
        """
        Should return the mean of the W distribution.
        at observed timepoints. The shaped of the returned tensor
        should be (n_timepoints, n_cell_types, n_cell_types)
        """
        raise NotImplementedError

    @property
    def F_mean(self):
        """
        Should return the mean of the F distribution.
        at observed timepoints. The shaped of the returned tensor
        should be (n_timepoints, n_cell_types, 1)
        """
        raise NotImplementedError

    def B_MEAN(self):
        """
        Should return the mean of the B distribution.
        at observed timepoints. The shaped of the returned tensor
        should be (n_timepoints, n_cell_types, 1)
        """
        raise NotImplementedError

    def _hypers_guide(self, hypers_init_vals, *args, **kwargs):
        """
        We use the same guide for the hyperparameters of every
        possible guide. Namely we use a MAP estimate of the
        hyperparameters.
        """
        init_lengthscale_w = torch.tensor(hypers_init_vals[names.LENGTHSCALE_W])
        w_lengthscale = pyro.param(
            names.LENGTHSCALE_W + "_param",
            init_lengthscale_w,
            constraint=constraints.positive,
        )
        w_lengthscale = pyro.sample(
            names.LENGTHSCALE_W, dist.Delta(w_lengthscale)
        )


@register_guide("multivariate_normal_factorized")
class MultivariateNormalFactorized(Guide):
    """
    Guide for the DIISCO algorithm implementing a factorized normal
    distribution via Cholesky parametrization. Described below.

    Details:
        We use q(W, f) = q(W)q(f) as the guide.
        q(W) = prod_{i=1}^{n_cell_types} prod_{j=1}^{n_cell_types} N(W_{ij})`
        $q(f) = prod_{i=1}^{n_cell_types} N(f_i)`
        Every q(W_{ij}) and q(f_i) are independent of each other but they are
        mulitvariate normal distributions coupled through time.
    """

    # Names of the parameters in the guide.
    F_MEAN = "f_mean"
    F_CHOLESKY_DIAG = "f_cholesky_diag"
    F_CHOLESKY_OFF_DIAG = "f_cholesky_off_diag"
    W_MEAN = "w_mean"
    W_CHOLESKY_DIAG = "w_cholesky_diag"
    W_CHOLESKY_OFF_DIAG = "w_cholesky_off_diag"
    B_MEAN = "b_mean"
    B_CHOLESKY_DIAG = "b_cholesky_diag"
    B_CHOLESKY_OFF_DIAG = "b_cholesky_off_diag"

    def __init__(
        self,
        n_cell_types,
        n_time_points,
        prior_f_mean,
        use_bias=False,
        lambda_matrix=None,
        hypers_init_vals=None,
        timepoints=None,
        proportions=None,
        *args,
        **kwargs
    ):
        # Check that the model has n_cell_types and n_timepoints
        # and raise an error if it doesn't.

        self.n_cell_types = n_cell_types
        self.n_timepoints = n_time_points
        self.prior_f_mean = prior_f_mean
        self.use_bias = use_bias
        self.hypers_init_vals = hypers_init_vals
        print("guide hyperparameters", self.hypers_init_vals)

        if (
            timepoints is not None
            and proportions is not None
            and lambda_matrix is not None
        ):
            init_W, init_B = self._init_params(
                timepoints, proportions, lambda_matrix
            )
            self.init_W = init_W
            self.init_B = init_B

    def __call__(
        self, timepoints: torch.Tensor, proportions: torch.Tensor = None
    ) -> None:
        """
        :param timepoints: Tensor of timepoints for the samples.
            shape: (n_timepoints, 1)
        :param proportions: Tensor of proportions for the samples.
            shape: (n_timepoints, n_cell_types)

        Note: This function is tightly coupled with the model and should have
        the same signature as the model's `forward` function.
        """
        # Call the guide for the hyperparameters.
        # This is an effectfull
        self._hypers_guide(hypers_init_vals=self.hypers_init_vals)

        n_cell_types = self.n_cell_types
        n_timepoints = self.n_timepoints

        # How many parameters we should have for the cholesky decomposition
        # of the covariance matrix for each coordinate through time.
        n_cholesky_params_w = int(n_timepoints * (n_timepoints + 1) / 2)
        n_cholesky_params_b = int(n_timepoints * (n_timepoints + 1) / 2)
        n_cholesky_params_f = int(n_timepoints * (n_timepoints + 1) / 2)

        n_cholesky_diag_params_w = n_timepoints
        n_cholesky_diag_params_b = n_timepoints
        n_cholesky_diag_params_f = n_timepoints

        cholesky_params_w = pyro.param(
            self.W_CHOLESKY_OFF_DIAG,
            torch.zeros(n_cell_types, n_cell_types, n_cholesky_params_w),
        )
        cholesky_params_b = pyro.param(
            self.B_CHOLESKY_OFF_DIAG,
            torch.zeros(n_cell_types, 1, n_cholesky_params_b),
        )
        cholesky_params_f = pyro.param(
            self.F_CHOLESKY_OFF_DIAG,
            torch.zeros(n_cell_types, n_cholesky_params_f),
        )

        cholesky_diag_params_w = pyro.param(
            self.W_CHOLESKY_DIAG,
            torch.abs(torch.randn(n_cholesky_diag_params_w)) * EPSILON,
            constraint=constraints.softplus_positive,
        )
        cholesky_diag_params_f = pyro.param(
            self.F_CHOLESKY_DIAG,
            torch.abs(torch.randn(n_cholesky_diag_params_f)) * EPSILON,
            constraint=constraints.softplus_positive,
        )
        cholesky_diag_params_b = pyro.param(
            self.B_CHOLESKY_DIAG,
            torch.abs(torch.randn(n_cholesky_diag_params_b)) * EPSILON,
            constraint=constraints.softplus_positive,
        )

        cholesky_w = vec_to_tril_matrix(cholesky_params_w)
        cholesky_f = vec_to_tril_matrix(cholesky_params_f)
        cholesky_b = vec_to_tril_matrix(cholesky_params_b)

        # Replace the diagonal with the cholesky diagonal parameters
        cholesky_w[
            :, :, torch.arange(n_timepoints), torch.arange(n_timepoints)
        ] = cholesky_diag_params_w
        cholesky_b[
            :, :, torch.arange(n_timepoints), torch.arange(n_timepoints)
        ] = cholesky_diag_params_b
        cholesky_f[
            :, torch.arange(n_timepoints), torch.arange(n_timepoints)
        ] = cholesky_diag_params_f

        covariances_w = torch.matmul(cholesky_w, cholesky_w.transpose(-1, -2))
        covariances_f = torch.matmul(cholesky_f, cholesky_f.transpose(-1, -2))
        covariances_b = torch.matmul(cholesky_b, cholesky_b.transpose(-1, -2))

        # Add small constant to the diagonal to make sure the
        # covariance matrix is positive definite
        covariances_f = covariances_f + torch.eye(n_timepoints) * EPSILON
        covariances_w = covariances_w + torch.eye(n_timepoints) * EPSILON
        covariances_b = covariances_b + torch.eye(n_timepoints) * EPSILON

        mean_w = pyro.param(
            self.W_MEAN,
            self.init_W
            + torch.randn(n_cell_types, n_cell_types, n_timepoints)
            * EPSILON**2,
        )
        mean_b = pyro.param(
            self.B_MEAN,
            self.init_B
            + torch.randn(n_cell_types, 1, n_timepoints) * EPSILON**2,
        )
        mean_f = pyro.param(self.F_MEAN, self.prior_f_mean).squeeze(-1)

        with pyro.plate("cell_types_outer_W", n_cell_types, dim=-2):
            with pyro.plate("cell_types_inner_W", n_cell_types, dim=-1):
                pyro.sample(
                    names.W, dist.MultivariateNormal(mean_w, covariances_w)
                )
        with pyro.plate("node_plate", n_cell_types, dim=-1):
            pyro.sample(names.F, dist.MultivariateNormal(mean_f, covariances_f))

        if self.use_bias:
            with pyro.plate("cell_types_outer_B", n_cell_types, dim=-2):
                with pyro.plate("single_inner_B", 1, dim=-1):
                    pyro.sample(
                        names.B, dist.MultivariateNormal(mean_b, covariances_b)
                    )

    def _init_params(self, timepoints, proportions, lambda_matrix):
        """
        Performs a simple linear regression to initialize the
        parameters of the parameters of the guide.

        Params:
        -------
        timepoints: torch.Tensor
            shape: (n_timepoints, 1)
        proportions: torch.Tensor
            shape (n_timepoints, n_cell_types)
        lambda_matrix: torch.Tensor
            shape (n_cell_types, n_cell_types)

        Returns:
        --------
        init_W: torch.Tensor
            shape: (n_cell_types, n_cell_types, n_timepoints)
        init_B: torch.Tensor
            shape: (n_cell_types, n_timepoints)
        """
        n_cell_types = proportions.shape[1]
        n_timepoints = proportions.shape[0]

        init_W = torch.zeros(n_cell_types, n_cell_types, n_timepoints)
        init_B = torch.zeros(n_cell_types, n_timepoints)

        for cell_type in range(n_cell_types):
            # Fit a linear regression to initialize the parameters
            # of the guide.

            # We zero out the cell types so that the model
            # cant use them.
            x = proportions * lambda_matrix[cell_type, :]
            y = proportions[:, cell_type]
            model = LinearRegression(fit_intercept=self.use_bias)
            model.fit(x, y)

            # We have collinearity and these values could be anything
            # so we just set them to zero.
            w = torch.tensor(model.coef_)
            w = w.flatten() * lambda_matrix[cell_type, :].flatten()
            if self.use_bias:
                b = torch.tensor(model.intercept_)

            init_W[cell_type, :, :] = w.unsqueeze(-1)
            init_B[cell_type, :] = b

        init_B.unsqueeze_(1)
        assert init_W.shape == (n_cell_types, n_cell_types, n_timepoints)
        assert init_B.shape == (n_cell_types, 1, n_timepoints)

        return init_W, init_B

    @property
    def W_mean(self):
        params = pyro.get_param_store()
        W_mean = params[self.W_MEAN]
        W_mean = W_mean.permute(2, 0, 1)
        assert W_mean.shape == (
            self.n_timepoints,
            self.n_cell_types,
            self.n_cell_types,
        )
        return W_mean

    @property
    def F_mean(self):
        params = pyro.get_param_store()
        F_mean = params[self.F_MEAN].permute(1, 0, 2)
        assert F_mean.shape == (self.n_timepoints, self.n_cell_types, 1)
        return F_mean

    @property
    def B_mean(self):
        params = pyro.get_param_store()
        B_mean = params[self.B_MEAN].permute(2, 0, 1)
        assert B_mean.shape == (self.n_timepoints, self.n_cell_types, 1)
        return B_mean


@register_guide("diagonal_normal")
class DiagonalNormal(Guide):
    """
    Guide for the DIISCO algorithm implementing a single
    normal distribution for each factor with diagonal covariance.
    """

    # Names of the parameters in the guide.
    F_MEAN = "f_mean"
    F_CHOLESKY_DIAG = "f_cholesky_diag"
    F_CHOLESKY_OFF_DIAG = "f_cholesky_off_diag"
    W_MEAN = "w_mean"
    W_CHOLESKY_DIAG = "w_cholesky_diag"
    W_CHOLESKY_OFF_DIAG = "w_cholesky_off_diag"
    B_MEAN = "b_mean"
    B_CHOLESKY_DIAG = "b_cholesky_diag"
    B_CHOLESKY_OFF_DIAG = "b_cholesky_off_diag"

    def __init__(
        self,
        n_cell_types,
        n_time_points,
        prior_f_mean,
        use_bias=False,
        lambda_matrix=None,
        hypers_init_vals=None,
        timepoints=None,
        proportions=None,
        *args,
        **kwargs
    ):
        # Check that the model has n_cell_types and n_timepoints
        # and raise an error if it doesn't.

        self.n_cell_types = n_cell_types
        self.n_timepoints = n_time_points
        self.prior_f_mean = prior_f_mean
        self.use_bias = use_bias
        self.hypers_init_vals = hypers_init_vals

        if (
            timepoints is not None
            and proportions is not None
            and lambda_matrix is not None
        ):
            init_W, init_B = self._init_params(
                timepoints, proportions, lambda_matrix
            )
            self.init_W = init_W
            self.init_B = init_B

    def __call__(
        self, timepoints: torch.Tensor, proportions: torch.Tensor = None
    ) -> None:
        """
        :param timepoints: Tensor of timepoints for the samples.
            shape: (n_timepoints, 1)
        :param proportions: Tensor of proportions for the samples.
            shape: (n_timepoints, n_cell_types)

        Note: This function is tightly coupled with the model and should have
        the same signature as the model's `forward` function.
        """
        # Call the guide for the hyperparameters.
        # This is an effectfull function.
        self._hypers_guide(hypers_init_vals=self.hypers_init_vals)

        n_cell_types = self.n_cell_types
        n_timepoints = self.n_timepoints

        n_cholesky_diag_params_w = n_timepoints * n_cell_types**2
        n_cholesky_diag_params_b = n_timepoints * n_cell_types
        n_cholesky_diag_params_f = n_timepoints * n_cell_types

        cholesky_diag_params_w = (
            pyro.param(
                self.W_CHOLESKY_DIAG,
                torch.abs(torch.randn(n_cholesky_diag_params_w)) * EPSILON,
                constraint=constraints.softplus_positive,
            )
            + EPSILON
        )
        cholesky_diag_params_w = cholesky_diag_params_w.reshape(
            n_cell_types, n_cell_types, n_timepoints
        )

        cholesky_diag_params_f = (
            pyro.param(
                self.F_CHOLESKY_DIAG,
                torch.abs(torch.randn(n_cholesky_diag_params_f)) * EPSILON,
                constraint=constraints.softplus_positive,
            )
            + EPSILON
        )
        cholesky_diag_params_f = cholesky_diag_params_f.reshape(
            n_cell_types, n_timepoints
        )

        cholesky_diag_params_b = (
            pyro.param(
                self.B_CHOLESKY_DIAG,
                torch.abs(torch.randn(n_cholesky_diag_params_b)) * EPSILON,
                constraint=constraints.softplus_positive,
            )
            + EPSILON
        )
        cholesky_diag_params_b = cholesky_diag_params_b.reshape(
            n_cell_types, 1, n_timepoints
        )

        mean_w = pyro.param(
            self.W_MEAN,
            self.init_W
            + torch.randn(n_cell_types, n_cell_types, n_timepoints)
            * EPSILON**2,
        )
        mean_b = pyro.param(
            self.B_MEAN,
            self.init_B
            + torch.randn(n_cell_types, 1, n_timepoints) * EPSILON**2,
        )
        mean_f = pyro.param(self.F_MEAN, self.prior_f_mean).squeeze(-1)

        with pyro.plate("cell_types_outer_W", n_cell_types, dim=-2):
            with pyro.plate("cell_types_inner_W", n_cell_types, dim=-1):
                W = pyro.sample(
                    names.W,
                    dist.Normal(mean_w, cholesky_diag_params_w).to_event(1),
                )
                assert W.shape == (n_cell_types, n_cell_types, n_timepoints)

        with pyro.plate("node_plate", n_cell_types, dim=-1):
            F = pyro.sample(
                names.F, dist.Normal(mean_f, cholesky_diag_params_f).to_event(1)
            )
            assert F.shape == (n_cell_types, n_timepoints)

        if self.use_bias:
            with pyro.plate("cell_types_outer_B", n_cell_types, dim=-2):
                with pyro.plate("single_inner_B", 1, dim=-1):
                    B = pyro.sample(
                        names.B,
                        dist.Normal(mean_b, cholesky_diag_params_b).to_event(1),
                    )
                assert B.shape == (n_cell_types, 1, n_timepoints)

    def _init_params(self, timepoints, proportions, lambda_matrix):
        """
        Performs a simple linear regression to initialize the
        parameters of the parameters of the guide.

        Params:
        -------
        timepoints: torch.Tensor
            shape: (n_timepoints, 1)
        proportions: torch.Tensor
            shape (n_timepoints, n_cell_types)
        lambda_matrix: torch.Tensor
            shape (n_cell_types, n_cell_types)

        Returns:
        --------
        init_W: torch.Tensor
            shape: (n_cell_types, n_cell_types, n_timepoints)
        init_B: torch.Tensor
            shape: (n_cell_types, n_timepoints)
        """
        n_cell_types = proportions.shape[1]
        n_timepoints = proportions.shape[0]

        init_W = torch.zeros(n_cell_types, n_cell_types, n_timepoints)
        init_B = torch.zeros(n_cell_types, n_timepoints)

        for cell_type in range(n_cell_types):
            # Fit a linear regression to initialize the parameters
            # of the guide.

            # We zero out the cell types so that the model
            # cant use them.
            x = proportions * lambda_matrix[cell_type, :]
            y = proportions[:, cell_type]
            model = LinearRegression(fit_intercept=self.use_bias)
            model.fit(x, y)

            # We have collinearity and these values could be anything
            # so we just set them to zero.
            w = torch.tensor(model.coef_)
            w = w.flatten() * lambda_matrix[cell_type, :].flatten()
            if self.use_bias:
                b = torch.tensor(model.intercept_)
                init_B[cell_type, :] = b

            init_W[cell_type, :, :] = w.unsqueeze(-1)


        init_B.unsqueeze_(1)
        assert init_W.shape == (n_cell_types, n_cell_types, n_timepoints)
        assert init_B.shape == (n_cell_types, 1, n_timepoints)

        init_W = torch.randn(n_cell_types, n_cell_types, n_timepoints) * 0.1
        init_B = torch.randn(n_cell_types, 1, n_timepoints) * 0.1

        return init_W, init_B

    @property
    def W_mean(self):
        params = pyro.get_param_store()
        W_mean = params[self.W_MEAN]
        W_mean = W_mean.permute(2, 0, 1)
        assert W_mean.shape == (
            self.n_timepoints,
            self.n_cell_types,
            self.n_cell_types,
        )
        return W_mean

    @property
    def F_mean(self):
        params = pyro.get_param_store()
        F_mean = params[self.F_MEAN].permute(1, 0, 2)
        assert F_mean.shape == (self.n_timepoints, self.n_cell_types, 1)
        return F_mean

    @property
    def B_mean(self):
        params = pyro.get_param_store()
        B_mean = params[self.B_MEAN].permute(2, 0, 1)
        assert B_mean.shape == (self.n_timepoints, self.n_cell_types, 1)
        return B_mean
